@@ -36,23 +36,17 @@
 #include "modules.h"
 
 typedef struct _PangoWin32Family       PangoWin32Family;
-typedef struct _PangoWin32SizeInfo     PangoWin32SizeInfo;
 
 struct _PangoWin32Family
 {
   PangoFontFamily parent_instance;
 
   char *family_name;
-  GSList *font_entries;
+  GSList *faces;
 
   gboolean is_monospace;
 };
 
-struct _PangoWin32SizeInfo
-{
-  GSList *logfontws;
-};
-
 #if !defined(NTM_PS_OPENTYPE)
 # define NTM_PS_OPENTYPE 0x20000
 #endif
@@ -96,6 +90,11 @@ static void       pango_win32_insert_font            (PangoWin32FontMap
 						      LOGFONTW                     *lfp,
 						      gboolean			    is_synthetic);
 
+static PangoWin32Family *pango_win32_get_font_family (PangoWin32FontMap            *win32fontmap,
+						      const char                   *family_name);
+
+static const char *pango_win32_face_get_face_name    (PangoFontFace *face);
+
 static PangoWin32FontMap *default_fontmap = NULL;
 
 G_DEFINE_TYPE (PangoWin32FontMap, _pango_win32_font_map, PANGO_TYPE_FONT_MAP)
@@ -196,7 +195,7 @@ pango_win32_enum_proc (LOGFONTW       *lfp,
 {
   LOGFONTW lf;
 
-  PING(("%S: %lu %lx", lfp->lfFaceName, fontType, metrics->ntmFlags));
+  PING (("%S: %lu %lx", lfp->lfFaceName, fontType, metrics->ntmFlags));
 
   if (fontType == TRUETYPE_FONTTYPE ||
       (_pango_win32_os_version_info.dwMajorVersion >= 5 &&
@@ -212,54 +211,161 @@ pango_win32_enum_proc (LOGFONTW       *lfp,
   return 1;
 }
 
-typedef struct _ItalicHelper
+static void
+synthesize_foreach (gpointer key,
+		    gpointer value,
+		    gpointer user_data)
 {
-  PangoWin32FontMap *fontmap;
-  GSList            *list;
-} ItalicHelper;
+  PangoWin32Family *win32family = value;
+  PangoWin32FontMap *win32fontmap = user_data;
+
+  enum { NORMAL, BOLDER, SLANTED };
+  PangoWin32Face *variant[4] = { NULL, NULL, NULL, NULL };
+
+  GSList *p;
+
+  LOGFONTW lf;
+
+  p = win32family->faces;
+  while (p)
+    {
+      PangoWin32Face *win32face = p->data;
+
+      /* Don't synthesize anything unless it's a monospace, serif, or sans font */
+      if (!((win32face->logfontw.lfPitchAndFamily & 0xF0) == FF_MODERN ||
+	    (win32face->logfontw.lfPitchAndFamily & 0xF0) == FF_ROMAN ||
+	    (win32face->logfontw.lfPitchAndFamily & 0xF0) == FF_SWISS))
+	return;
+
+      if (pango_font_description_get_weight (win32face->description) == PANGO_WEIGHT_NORMAL &&
+	  pango_font_description_get_style (win32face->description) == PANGO_STYLE_NORMAL)
+	variant[NORMAL] = win32face;
+
+      if (pango_font_description_get_weight (win32face->description) > PANGO_WEIGHT_NORMAL &&
+	  pango_font_description_get_style (win32face->description) == PANGO_STYLE_NORMAL)
+	variant[BOLDER] = win32face;
+
+      if (pango_font_description_get_weight (win32face->description) == PANGO_WEIGHT_NORMAL &&
+	  pango_font_description_get_style (win32face->description) >= PANGO_STYLE_OBLIQUE)
+	variant[SLANTED] = win32face;
+
+      if (pango_font_description_get_weight (win32face->description) > PANGO_WEIGHT_NORMAL &&
+	  pango_font_description_get_style (win32face->description) >= PANGO_STYLE_OBLIQUE)
+	variant[BOLDER+SLANTED] = win32face;
+
+      p = p->next;
+    }
+
+  if (variant[NORMAL] != NULL && variant[BOLDER] == NULL)
+    {
+      lf = variant[NORMAL]->logfontw;
+      lf.lfWeight = FW_BOLD;
+      
+      pango_win32_insert_font (win32fontmap, &lf, TRUE);
+    }
+
+  if (variant[NORMAL] != NULL && variant[SLANTED] == NULL)
+    {
+      lf = variant[NORMAL]->logfontw;
+      lf.lfItalic = 255;
+      
+      pango_win32_insert_font (win32fontmap, &lf, TRUE);
+    }
+
+  if (variant[NORMAL] != NULL &&
+      variant[BOLDER+SLANTED] == NULL)
+    {
+      lf = variant[NORMAL]->logfontw;
+      lf.lfWeight = FW_BOLD;
+      lf.lfItalic = 255;
+
+      pango_win32_insert_font (win32fontmap, &lf, TRUE);
+    }
+  else if (variant[BOLDER] != NULL &&
+	   variant[BOLDER+SLANTED] == NULL)
+    {
+      lf = variant[BOLDER]->logfontw;
+      lf.lfItalic = 255;
+
+      pango_win32_insert_font (win32fontmap, &lf, TRUE);
+    }
+  else if (variant[SLANTED] != NULL &&
+	   variant[BOLDER+SLANTED] == NULL)
+    {
+      lf = variant[SLANTED]->logfontw;
+      lf.lfWeight = FW_BOLD;
+
+      pango_win32_insert_font (win32fontmap, &lf, TRUE);
+    }
+}
 
 static void
-ensure_italic (gpointer key,
-	       gpointer value,
-	       gpointer user_data)
+create_standard_family (PangoWin32FontMap *win32fontmap,
+			const char        *standard_family_name)
 {
-  ItalicHelper *helper = (ItalicHelper *)user_data;
-  PangoWin32SizeInfo *sip = (PangoWin32SizeInfo *) value;
-  GSList *list = sip->logfontws;
+  int i;
+  int n_aliases;
+  char **aliases;
 
-  while (list)
+  pango_lookup_aliases (standard_family_name, &aliases, &n_aliases);
+  for (i = 0; i < n_aliases; i++)
     {
-      LOGFONTW *lfp = (LOGFONTW *) list->data;
-      PING(("%S it=%d wt=%ld", lfp->lfFaceName, lfp->lfItalic, lfp->lfWeight));
-      if (!lfp->lfItalic)
+      PangoWin32Family *existing_family = g_hash_table_lookup (win32fontmap->families, aliases[i]);
+
+      if (existing_family)
 	{
-	  /* we have a non italic variant, look if there is an italic */
-	  LOGFONTW logfontw = *lfp;
-	  logfontw.lfItalic = 1;
-	  sip = (PangoWin32SizeInfo *) g_hash_table_lookup (helper->fontmap->size_infos, &logfontw);
-	  if (!sip)
+	  PangoWin32Family *new_family = pango_win32_get_font_family (win32fontmap, standard_family_name);
+	  GSList *p = existing_family->faces;
+
+	  new_family->is_monospace = existing_family->is_monospace;
+
+	  while (p)
 	    {
-	      /* remember the non italic variant to be added later as italic */
-	      PING(("synthesizing italic"));
-	      helper->list = g_slist_append (helper->list, lfp);
+	      const PangoWin32Face *old_face = p->data;
+	      PangoWin32Face *new_face = g_object_new (PANGO_WIN32_TYPE_FACE, NULL);
+	      int j;
+
+	      new_face->logfontw = old_face->logfontw;
+	      new_face->description = pango_font_description_copy_static (old_face->description);
+	      pango_font_description_set_family_static (new_face->description, standard_family_name);
+
+	      for (j = 0; j < PANGO_WIN32_N_COVERAGES; j++)
+		{
+		  if (old_face->coverages[j] != NULL)
+		    new_face->coverages[j] = pango_coverage_ref (old_face->coverages[j]);
+		  else
+		    new_face->coverages[j] = NULL;
+		}
+
+	      new_face->is_synthetic = TRUE;
+
+	      new_face->has_cmap = old_face->has_cmap;
+	      new_face->cmap_format = old_face->cmap_format;
+	      new_face->cmap = old_face->cmap;
+
+	      new_face->cached_fonts = NULL;
+
+	      new_family->faces = g_slist_append (new_family->faces, new_face);
+
+	      p = p->next;
 	    }
+	  return;
 	}
-      list = list->next;
     }
+  /* XXX What to do if none of the members of aliases for standard_family_name
+   * exists on this machine?
+   */
 }
 
 static void
 _pango_win32_font_map_init (PangoWin32FontMap *win32fontmap)
 {
   LOGFONTW logfont;
-  ItalicHelper helper = { win32fontmap, NULL };
-  GSList *list;
 
   win32fontmap->families = g_hash_table_new ((GHashFunc) case_insensitive_str_hash,
 					     (GEqualFunc) case_insensitive_str_equal);
-  win32fontmap->size_infos =
+  win32fontmap->fonts =
     g_hash_table_new ((GHashFunc) logfontw_nosize_hash, (GEqualFunc) logfontw_nosize_equal);
-  win32fontmap->n_fonts = 0;
 
   win32fontmap->font_cache = pango_win32_font_cache_new ();
   win32fontmap->freed_fonts = g_queue_new ();
@@ -270,19 +376,12 @@ _pango_win32_font_map_init (PangoWin32FontMap *win32fontmap)
 		       (FONTENUMPROCW) pango_win32_enum_proc,
 		       (LPARAM) win32fontmap, 0);
 
-  /* Create synthetic italic entries */
-  g_hash_table_foreach (win32fontmap->size_infos, ensure_italic, &helper);
+  g_hash_table_foreach (win32fontmap->families, synthesize_foreach, win32fontmap);
 
-  /* Can't modify while iterating */
-  list = helper.list;
-  while (list)
-    {
-      LOGFONTW logfontw = *((LOGFONTW *)list->data);
-      logfontw.lfItalic = 1;
-      pango_win32_insert_font (win32fontmap, &logfontw, TRUE);
-      list = list->next;
-    }
-  g_slist_free (helper.list);
+  /* Create synthetic "Sans", "Serif" and "Monospace" families */
+  create_standard_family (win32fontmap, "Sans");
+  create_standard_family (win32fontmap, "Serif");
+  create_standard_family (win32fontmap, "Monospace");
 
   win32fontmap->resolution = (PANGO_SCALE / (double) GetDeviceCaps (_pango_win32_hdc, LOGPIXELSY)) * 72.0;
 }
@@ -368,28 +467,40 @@ pango_win32_family_list_faces (PangoFontFamily  *family,
 			       int              *n_faces)
 {
   PangoWin32Family *win32family = PANGO_WIN32_FAMILY (family);
+  GSList *p;
+  int n;
+
+  p = win32family->faces;
+  n = 0;
+  while (p)
+    {
+      n++;
+      p = p->next;
+    }
 
-  *n_faces = g_slist_length (win32family->font_entries);
   if (faces)
     {
-      GSList *tmp_list;
-      int i = 0;
+      int i;
 
-      *faces = g_new (PangoFontFace *, *n_faces);
+      *faces = g_new (PangoFontFace *, n);
 
-      tmp_list = win32family->font_entries;
-      while (tmp_list)
+      p = win32family->faces;
+      i = 0;
+      while (p)
 	{
-	  (*faces)[i++] = tmp_list->data;
-	  tmp_list = tmp_list->next;
+	  (*faces)[i++] = p->data;
+	  p = p->next;
 	}
     }
+  if (n_faces)
+    *n_faces = n;
 }
 
 static const char *
 pango_win32_family_get_name (PangoFontFamily  *family)
 {
   PangoWin32Family *win32family = PANGO_WIN32_FAMILY (family);
+
   return win32family->family_name;
 }
 
@@ -490,7 +601,7 @@ pango_win32_get_font_family (PangoWin32FontMap *win32fontmap,
     {
       win32family = g_object_new (PANGO_WIN32_TYPE_FAMILY, NULL);
       win32family->family_name = g_strdup (family_name);
-      win32family->font_entries = NULL;
+      win32family->faces = NULL;
 
       g_hash_table_insert (win32fontmap->families, win32family->family_name, win32family);
     }
@@ -510,7 +621,7 @@ pango_win32_font_map_load_font (PangoFontMap               *fontmap,
 
   g_return_val_if_fail (description != NULL, NULL);
 
-  PING(("name=%s", pango_font_description_get_family (description)));
+  PING (("name=%s", pango_font_description_get_family (description)));
 
   win32family = g_hash_table_lookup (win32fontmap->families,
 				     pango_font_description_get_family (description));
@@ -519,7 +630,7 @@ pango_win32_font_map_load_font (PangoFontMap               *fontmap,
       PangoWin32Face *best_match = NULL;
 
       PING (("got win32family"));
-      tmp_list = win32family->font_entries;
+      tmp_list = win32family->faces;
       while (tmp_list)
 	{
 	  PangoWin32Face *face = tmp_list->data;
@@ -538,7 +649,7 @@ pango_win32_font_map_load_font (PangoFontMap               *fontmap,
 									   description);
 	/* TODO: Handle the case that result == NULL. */
       else
-	PING(("no best match!"));
+	PING (("no best match!"));
     }
 
   return result;
@@ -579,7 +690,7 @@ pango_win32_font_map_real_find_font (PangoWin32FontMap          *win32fontmap,
   if (pango_font_description_get_size_is_absolute (description))
     size = (int) 0.5 + (size * win32fontmap->resolution) / PANGO_SCALE;
 
-  PING(("got best match:%S size=%d",face->logfontw.lfFaceName,size));
+  PING (("got best match:%S size=%d",face->logfontw.lfFaceName,size));
 
   while (tmp_list)
     {
@@ -660,8 +771,8 @@ get_family_nameA (const LOGFONTA *lfp)
       if ((record.name_id != 1 && record.name_id != 16) || record.string_length <= 0)
 	continue;
 
-      PING(("platform:%d encoding:%d language:%04x name_id:%d",
-	    record.platform_id, record.encoding_id, record.language_id, record.name_id));
+      PING (("platform:%d encoding:%d language:%04x name_id:%d",
+	     record.platform_id, record.encoding_id, record.language_id, record.name_id));
 
       if (record.platform_id == APPLE_UNICODE_PLATFORM_ID ||
 	  record.platform_id == ISO_PLATFORM_ID)
@@ -716,7 +827,7 @@ get_family_nameA (const LOGFONTA *lfp)
     goto fail2;
   g_free (string);
 
-  PING(("%s", name));
+  PING (("%s", name));
 
   SelectObject (_pango_win32_hdc, oldhfont);
   DeleteObject (hfont);
@@ -782,8 +893,10 @@ pango_win32_font_description_from_logfont (const LOGFONT *lfp)
     weight = PANGO_WEIGHT_ULTRALIGHT;
   else if (lfp->lfWeight <= (FW_LIGHT + FW_NORMAL) / 2)
     weight = PANGO_WEIGHT_LIGHT;
-  else if (lfp->lfWeight <= (FW_NORMAL + FW_BOLD) / 2)
+  else if (lfp->lfWeight <= (FW_NORMAL + FW_SEMIBOLD) / 2)
     weight = PANGO_WEIGHT_NORMAL;
+  else if (lfp->lfWeight <= (FW_SEMIBOLD + FW_BOLD) / 2)
+    weight = PANGO_WEIGHT_SEMIBOLD;
   else if (lfp->lfWeight <= (FW_BOLD + FW_ULTRABOLD) / 2)
     weight = PANGO_WEIGHT_BOLD;
   else if (lfp->lfWeight <= (FW_ULTRABOLD + FW_HEAVY) / 2)
@@ -856,8 +969,8 @@ get_family_nameW (const LOGFONTW *lfp)
       if ((record.name_id != 1 && record.name_id != 16) || record.string_length <= 0)
 	continue;
 
-      PING(("platform:%d encoding:%d language:%04x name_id:%d",
-	    record.platform_id, record.encoding_id, record.language_id, record.name_id));
+      PING (("platform:%d encoding:%d language:%04x name_id:%d",
+	     record.platform_id, record.encoding_id, record.language_id, record.name_id));
 
       if (record.platform_id == APPLE_UNICODE_PLATFORM_ID ||
 	  record.platform_id == ISO_PLATFORM_ID)
@@ -912,7 +1025,7 @@ get_family_nameW (const LOGFONTW *lfp)
     goto fail2;
   g_free (string);
 
-  PING(("%s", name));
+  PING (("%s", name));
 
   SelectObject (_pango_win32_hdc, oldhfont);
   DeleteObject (hfont);
@@ -956,15 +1069,71 @@ pango_win32_font_description_from_logfontw (const LOGFONTW *lfp)
   gchar *family;
   PangoStyle style;
   PangoVariant variant;
-  PangoWeight weight;
+  PangoWeight weight, name_weight;
   PangoStretch stretch;
 
+  static const struct {
+    const char *marker;
+    int marker_len;
+    int remove_len;
+    PangoWeight weight;
+  } weight_names[] = {
+#define ENTRY(n, s) ENTRY2 (n, sizeof (#n) - 1, s)
+#define ENTRY2(n, l, s) ENTRY3 (n, l, l, s)
+#define ENTRY3(n, marker_len, remove_len, s) { #n, marker_len, remove_len, PANGO_WEIGHT_##s }
+    ENTRY (Ultra Light, ULTRALIGHT),
+    ENTRY (UltraLight, ULTRALIGHT),
+    ENTRY (Light, LIGHT),
+    ENTRY (Medium, NORMAL),
+    ENTRY (Demi Bold, SEMIBOLD),
+    ENTRY (Demi, SEMIBOLD),
+    ENTRY (Ultra Bold, ULTRABOLD),
+    ENTRY (Extra Bold, ULTRABOLD),
+    ENTRY (SemiBold, SEMIBOLD),
+    ENTRY (DemiBold, SEMIBOLD),
+    ENTRY (UltraBold, ULTRABOLD),
+    ENTRY (ExtraBold, ULTRABOLD),
+    ENTRY (Bold, BOLD),
+    ENTRY (Heavy, HEAVY),
+    ENTRY (Black, HEAVY),
+#undef ENTRY
+#undef ENTRY2
+#undef ENTRY3
+  };
+
+  static const struct {
+    const char *marker;
+    int marker_len;
+    PangoStretch stretch;
+  } stretch_names[] = {
+#define ENTRY(n, s) { #n, sizeof (#n) - 1, PANGO_STRETCH_##s }
+    ENTRY (Ext Condensed, EXTRA_CONDENSED),
+    ENTRY (Extra Condensed, EXTRA_CONDENSED),
+    ENTRY (UltraCondensed, ULTRA_CONDENSED),
+    ENTRY (ExtraCondensed, EXTRA_CONDENSED),
+    ENTRY (Condensed, CONDENSED),
+    ENTRY (Cond, CONDENSED),
+    ENTRY (Narrow, CONDENSED),
+    ENTRY (Ext Expanded, EXTRA_EXPANDED),
+    ENTRY (Extra Expanded, EXTRA_EXPANDED),
+    ENTRY (Ultra Expanded, ULTRA_EXPANDED),
+    ENTRY (ExtraExpanded, EXTRA_EXPANDED),
+    ENTRY (UltraExpanded, ULTRA_EXPANDED),
+    ENTRY (Expanded, EXPANDED),
+#undef ENTRY
+  };
+
+  int i;
+  char *p;
+
   family = get_family_nameW (lfp);
 
-  if (!lfp->lfItalic)
-    style = PANGO_STYLE_NORMAL;
-  else
+  if ((lfp->lfPitchAndFamily & 0xF0) == FF_ROMAN && lfp->lfItalic)
     style = PANGO_STYLE_ITALIC;
+  else if (lfp->lfItalic)
+    style = PANGO_STYLE_OBLIQUE;
+  else
+    style = PANGO_STYLE_NORMAL;
 
   variant = PANGO_VARIANT_NORMAL;
 
@@ -978,8 +1147,10 @@ pango_win32_font_description_from_logfontw (const LOGFONTW *lfp)
     weight = PANGO_WEIGHT_ULTRALIGHT;
   else if (lfp->lfWeight <= (FW_LIGHT + FW_NORMAL) / 2)
     weight = PANGO_WEIGHT_LIGHT;
-  else if (lfp->lfWeight <= (FW_NORMAL + FW_BOLD) / 2)
+  else if (lfp->lfWeight <= (FW_NORMAL + FW_SEMIBOLD) / 2)
     weight = PANGO_WEIGHT_NORMAL;
+  else if (lfp->lfWeight <= (FW_SEMIBOLD + FW_BOLD) / 2)
+    weight = PANGO_WEIGHT_SEMIBOLD;
   else if (lfp->lfWeight <= (FW_BOLD + FW_ULTRABOLD) / 2)
     weight = PANGO_WEIGHT_BOLD;
   else if (lfp->lfWeight <= (FW_ULTRABOLD + FW_HEAVY) / 2)
@@ -987,9 +1158,51 @@ pango_win32_font_description_from_logfontw (const LOGFONTW *lfp)
   else
     weight = PANGO_WEIGHT_HEAVY;
 
-  /* XXX No idea how to figure out the stretch */
+  name_weight = 0;
+
+  p = family;
+  while ((p = strchr (p, ' ')) != NULL)
+    {
+      for (i = 0; i < G_N_ELEMENTS (weight_names); i++)
+	{
+	  if (g_ascii_strncasecmp (p + 1, weight_names[i].marker, weight_names[i].marker_len) == 0 &&
+	      (p[1 + weight_names[i].marker_len] == '\0' ||
+	       p[1 + weight_names[i].marker_len] == ' '))
+	    {
+	      strcpy (p, p + 1 + weight_names[i].remove_len);
+	      name_weight = weight_names[i].weight;
+	      break;
+	    }
+	}
+      if (i < G_N_ELEMENTS (weight_names))
+	break;
+      p++;
+    }
+
+  if (weight == PANGO_WEIGHT_NORMAL && name_weight > 0)
+    weight = name_weight;
+
   stretch = PANGO_STRETCH_NORMAL;
 
+  p = family;
+  while ((p = strchr (p, ' ')) != NULL)
+    {
+      for (i = 0; i < G_N_ELEMENTS (stretch_names); i++)
+	{
+	  if (g_ascii_strncasecmp (p + 1, stretch_names[i].marker, stretch_names[i].marker_len) == 0 &&
+	      (p[1 + stretch_names[i].marker_len] == '\0' ||
+	       p[1 + stretch_names[i].marker_len] == ' '))
+	    {
+	      strcpy (p, p + 1 + stretch_names[i].marker_len);
+	      stretch = stretch_names[i].stretch;
+	      break;
+	    }
+	}
+      if (i < G_N_ELEMENTS (stretch_names))
+	break;
+      p++;
+    }
+
   description = pango_font_description_new ();
   pango_font_description_set_family (description, family);
   g_free(family);
@@ -1001,12 +1214,61 @@ pango_win32_font_description_from_logfontw (const LOGFONTW *lfp)
   return description;
 }
 
+static char *
+charset_name (int charset)
+{
+  static char num[10];
+
+  switch (charset)
+    {
+#define CASE(x) case x##_CHARSET: return #x
+      CASE (ANSI);
+      CASE (DEFAULT);
+      CASE (SYMBOL);
+      CASE (SHIFTJIS);
+      CASE (HANGUL);
+      CASE (GB2312);
+      CASE (CHINESEBIG5);
+      CASE (GREEK);
+      CASE (TURKISH);
+      CASE (HEBREW);
+      CASE (ARABIC);
+      CASE (BALTIC);
+      CASE (RUSSIAN);
+      CASE (THAI);
+      CASE (EASTEUROPE);
+      CASE (OEM);
+      CASE (JOHAB);
+      CASE (VIETNAMESE);
+      CASE (MAC);
+#undef CASE
+    default:
+      sprintf (num, "%d", charset);
+      return num;
+    }
+}
+
+static char *
+ff_name (int ff)
+{
+  static char num[10];
+
+  switch (ff)
+    {
+#define CASE(x) case FF_##x: return #x
+      CASE (DECORATIVE);
+      CASE (DONTCARE);
+      CASE (MODERN);
+      CASE (ROMAN);
+      CASE (SCRIPT);
+      CASE (SWISS);
+#undef CASE
+    default:
+      sprintf (num, "%d", ff);
+      return num;
+    }
+}
 
-/* This inserts the given font into the size_infos table. If a
- * SizeInfo already exists with the same typeface name, italicness and
- * weight, then the font is added to the SizeInfo's list, else a
- * new SizeInfo is created and inserted in the table.
- */
 static void
 pango_win32_insert_font (PangoWin32FontMap *win32fontmap,
 			 LOGFONTW          *lfp,
@@ -1014,14 +1276,18 @@ pango_win32_insert_font (PangoWin32FontMap *win32fontmap,
 {
   LOGFONTW *lfp2 = NULL;
   PangoFontDescription *description;
-  PangoWin32Family *font_family;
+  PangoWin32Family *win32family;
   PangoWin32Face *win32face;
-  PangoWin32SizeInfo *size_info;
-  GSList *tmp_list;
   gint i;
-  gchar *p;
 
-  PING(("face=%S,charset=%d,it=%d,wt=%ld,ht=%ld",lfp->lfFaceName,lfp->lfCharSet,lfp->lfItalic,lfp->lfWeight,lfp->lfHeight));
+  PING (("face=%S,charset=%s,it=%s,wt=%ld,ht=%ld,ff=%s%s",
+	 lfp->lfFaceName,
+	 charset_name (lfp->lfCharSet),
+	 lfp->lfItalic ? "yes" : "no",
+	 lfp->lfWeight,
+	 lfp->lfHeight,
+	 ff_name (lfp->lfPitchAndFamily & 0xF0),
+	 is_synthetic ? " synthetic" : ""));
 
   /* Ignore Symbol fonts (which don't have any Unicode mapping
    * table). We could also be fancy and use the PostScript glyph name
@@ -1031,55 +1297,16 @@ pango_win32_insert_font (PangoWin32FontMap *win32fontmap,
   if (lfp->lfCharSet == SYMBOL_CHARSET)
     return;
 
-  /* First insert the LOGFONTW into the list of LOGFONTWs for the
-   * typeface name, italicness and weight.
-   */
-  size_info = g_hash_table_lookup (win32fontmap->size_infos, lfp);
-  if (!size_info)
+  if (g_hash_table_lookup (win32fontmap->fonts, lfp))
     {
-      PING(("SizeInfo not found"));
-      size_info = g_new (PangoWin32SizeInfo, 1);
-      size_info->logfontws = NULL;
-
-      lfp2 = g_new (LOGFONTW, 1);
-      *lfp2 = *lfp;
-      g_hash_table_insert (win32fontmap->size_infos, lfp2, size_info);
-    }
-  else
-    {
-      /* Don't store LOGFONTWs that differ only in charset
-       */
-      tmp_list = size_info->logfontws;
-      while (tmp_list)
-	{
-	  LOGFONTW *rover = tmp_list->data;
-
-	  /* We know that lfWeight, lfItalic and lfFaceName match.  We
-	   * don't check lfHeight and lfWidth, those are used
-	   * when creating a font.
-	   */
-	  if (rover->lfEscapement == lfp->lfEscapement &&
-	      rover->lfOrientation == lfp->lfOrientation &&
-	      rover->lfUnderline == lfp->lfUnderline &&
-	      rover->lfStrikeOut == lfp->lfStrikeOut)
-	    {
-	      PING(("already have it"));
-	      return;
-	    }
-
-	  tmp_list = tmp_list->next;
-	}
+      PING (("already have it"));
+      return;
     }
 
-  if (lfp2 == NULL)
-    {
-      lfp2 = g_new (LOGFONTW, 1);
-      *lfp2 = *lfp;
-    }
-
-  size_info->logfontws = g_slist_prepend (size_info->logfontws, lfp2);
-
-  PING(("g_slist_length(size_info->logfontws)=%d", g_slist_length(size_info->logfontws)));
+  PING (("not found"));
+  lfp2 = g_new (LOGFONTW, 1);
+  *lfp2 = *lfp;
+  g_hash_table_insert (win32fontmap->fonts, lfp2, lfp2);
 
   description = pango_win32_font_description_from_logfontw (lfp2);
 
@@ -1094,6 +1321,8 @@ pango_win32_insert_font (PangoWin32FontMap *win32fontmap,
 
   win32face = g_object_new (PANGO_WIN32_TYPE_FACE, NULL);
 
+  PING (("win32face created: %p for %S", win32face, lfp->lfFaceName));
+
   win32face->logfontw = *lfp;
   win32face->description = description;
 
@@ -1108,58 +1337,16 @@ pango_win32_insert_font (PangoWin32FontMap *win32fontmap,
 
   win32face->cached_fonts = NULL;
 
-  font_family =
+  win32family =
     pango_win32_get_font_family (win32fontmap,
 				 pango_font_description_get_family (win32face->description));
-  font_family->font_entries = g_slist_append (font_family->font_entries, win32face);
-  PING(("g_slist_length(font_family->font_entries)=%d", g_slist_length(font_family->font_entries)));
+  if ((lfp->lfPitchAndFamily & 0xF0) == FF_MODERN)
+    win32family->is_monospace = TRUE;
+    
+  win32family->faces = g_slist_append (win32family->faces, win32face);
 
-  win32fontmap->n_fonts++;
-
-#if 1 /* Thought pango.aliases would make this code unnecessary, but no. */
-  /*
-   * There are magic family names coming from the X implementation.
-   * They can be simply mapped to lfPitchAndFamily flag of the logfont
-   * struct. These additional entries should probably only be references
-   * to the respective entry created above. Thy are simply using the
-   * same entry at the moment and it isn't crashing on g_free () ???
-   * Maybe a memory leak ...
-   */
-  switch (lfp->lfPitchAndFamily & 0xF0)
-    {
-    case FF_MODERN : /* monospace */
-      PING(("monospace"));
-      font_family->is_monospace = TRUE; /* modify before reuse */
-      font_family = pango_win32_get_font_family (win32fontmap, "monospace");
-      font_family->font_entries = g_slist_append (font_family->font_entries, win32face);
-      win32fontmap->n_fonts++;
-      break;
-    case FF_ROMAN : /* serif */
-      PING(("serif"));
-      font_family = pango_win32_get_font_family (win32fontmap, "serif");
-      font_family->font_entries = g_slist_append (font_family->font_entries, win32face);
-      win32fontmap->n_fonts++;
-      break;
-    case  FF_SWISS  : /* sans */
-      PING(("sans"));
-      font_family = pango_win32_get_font_family (win32fontmap, "sans");
-      font_family->font_entries = g_slist_append (font_family->font_entries, win32face);
-      win32fontmap->n_fonts++;
-      break;
-    }
-
-  /* Some other magic names */
-
-  /* Recognize just "courier" for "courier new" */
-  p = g_utf16_to_utf8 (win32face->logfontw.lfFaceName, -1, NULL, NULL, NULL);
-  if (p && g_ascii_strcasecmp (p, "courier new") == 0)
-    {
-      font_family = pango_win32_get_font_family (win32fontmap, "courier");
-      font_family->font_entries = g_slist_append (font_family->font_entries, win32face);
-      win32fontmap->n_fonts++;
-    }
-  g_free (p);
-#endif
+  PING (("name=%s, length(faces)=%d",
+	 win32family->family_name, g_slist_length (win32family->faces)));
 }
 
 /* Given a LOGFONTW and size, make a matching LOGFONTW corresponding to
@@ -1172,54 +1359,25 @@ _pango_win32_make_matching_logfontw (PangoFontMap   *fontmap,
 				     LOGFONTW       *out)
 {
   PangoWin32FontMap *win32fontmap;
-  GSList *tmp_list;
-  PangoWin32SizeInfo *size_info;
-  LOGFONTW *closest_match = NULL;
-  gint match_distance = 0;
+  LOGFONTW *match;
 
-  PING(("lfp.face=%S,wt=%ld,ht=%ld,size:%d",lfp->lfFaceName,lfp->lfWeight,lfp->lfHeight,size));
+  PING (("lfp.face=%S,wt=%ld,ht=%ld,size:%d",
+	 lfp->lfFaceName, lfp->lfWeight, lfp->lfHeight, size));
   win32fontmap = PANGO_WIN32_FONT_MAP (fontmap);
 
-  size_info = g_hash_table_lookup (win32fontmap->size_infos, lfp);
+  match = g_hash_table_lookup (win32fontmap->fonts, lfp);
 
-  if (!size_info)
+  if (!match)
     {
-      PING(("SizeInfo not found"));
+      PING (("not found"));
       return;
     }
 
-  tmp_list = size_info->logfontws;
-  while (tmp_list)
-    {
-      LOGFONTW *tmp_logfontw = tmp_list->data;
-      int font_size = abs (tmp_logfontw->lfHeight);
+  /* OK, we have a match; let's modify it to fit this size */
 
-      if (size != -1)
-	{
-	  int new_distance = (font_size == 0) ? 0 : abs (font_size - size);
-
-	  if (!closest_match ||
-	      new_distance < match_distance ||
-	      (new_distance < PANGO_SCALE && font_size != 0))
-	    {
-	      closest_match = tmp_logfontw;
-	      match_distance = new_distance;
-	    }
-	}
-
-      tmp_list = tmp_list->next;
-    }
-
-  if (closest_match)
-    {
-      /* OK, we have a match; let's modify it to fit this size */
-
-      *out = *closest_match;
-      out->lfHeight = -(int)((double)size / win32fontmap->resolution + 0.5);
-      out->lfWidth = 0;
-    }
-  else
-    *out = *lfp; /* Whatever. We need to pass something... */
+  *out = *match;
+  out->lfHeight = -(int)((double)size / win32fontmap->resolution + 0.5);
+  out->lfWidth = 0;
 }
 
 static PangoFontDescription *
